from qibo import gates
from qibo.config import log, raise_error


def find_connected_qubit(qubits, queue, hardware_qubits):
    """Helper method for :meth:`qibolab.transpilers.fix_connecivity`.

    Finds which qubit should be mapped to hardware middle qubit
    by looking at the two-qubit gates that follow.
    """
    possible_qubits = set(qubits)
    for next_gate in queue:
        if len(next_gate.qubits) == 2:
            possible_qubits &= {hardware_qubits.index(q) for q in next_gate.qubits}
            if not possible_qubits:
                # freedom of choice
                return qubits[0]
            elif len(possible_qubits) == 1:
                return possible_qubits.pop()
    # freedom of choice
    return qubits[0]


def fix_connectivity(circuit, middle_qubit=2):
    """Transforms an arbitrary circuit to one that can be executed on hardware.

    This method produces a circuit that respects the following connectivity:
          q
          |
     q -- q -- q
          |
          q
    by adding SWAP gates when needed.
    It does not translate gates to native.

    Args:
        circuit (qibo.models.Circuit): The original Qibo circuit to transform.
            This circuit must contain up to two-qubit gates.
        middle_qubit (int): Hardware middle qubit.

    Returns:
        new (qibo.models.Circuit): Qibo circuit that performs the same operation
            as the original but respects the hardware connectivity.
        hardware_qubits (list): List that maps logical to hardware qubits.
            This is required for transforming final measurements.
    """
    # TODO: Change this to a more lightweight form that takes a list of pairs
    # instead of the whole circuit.

    # find the number of qubits for hardware circuit
    if circuit.nqubits == 1:
        nqubits = 1
    else:
        nqubits = max(circuit.nqubits, middle_qubit + 1)
    # new circuit object that will be compatible with hardware connectivity
    new = circuit.__class__(nqubits)
    # list to maps logical to hardware qubits
    hardware_qubits = list(range(nqubits))

    # find initial qubit mapping
    for i, gate in enumerate(circuit.queue):
        if len(gate.qubits) == 2:
            if middle_qubit not in gate.qubits:
                new_middle = find_connected_qubit(gate.qubits, circuit.queue[i + 1 :], hardware_qubits)
                hardware_qubits[middle_qubit], hardware_qubits[new_middle] = (
                    hardware_qubits[new_middle],
                    hardware_qubits[middle_qubit],
                )
            break

    # the first SWAP is not needed as it can be applied via virtual mapping
    add_swap = False
    for i, gate in enumerate(circuit.queue):
        # map gate qubits to hardware
        qubits = tuple(hardware_qubits.index(q) for q in gate.qubits)
        if isinstance(gate, gates.M):
            new_gate = gates.M(*qubits, **gate.init_kwargs)
            new_gate.result = gate.result
            new.add(new_gate)
            continue

        if len(qubits) > 2:
            raise_error(
                NotImplementedError,
                "Transpiler does not support gates targeting more than two-qubits.",
            )

        elif len(qubits) == 2 and middle_qubit not in qubits:
            # find which qubit should be moved to 0
            new_middle = find_connected_qubit(qubits, circuit.queue[i + 1 :], hardware_qubits)
            # update hardware qubits according to the swap
            hardware_qubits[middle_qubit], hardware_qubits[new_middle] = (
                hardware_qubits[new_middle],
                hardware_qubits[middle_qubit],
            )
            if add_swap:
                new.add(gates.SWAP(middle_qubit, new_middle))
            # update gate qubits according to the new swap
            qubits = tuple(hardware_qubits.index(q) for q in gate.qubits)

        # add gate to the hardware circuit
        if isinstance(gate, gates.Unitary):
            # gates.Unitary requires matrix as first argument
            from qibo.backends import NumpyBackend

            backend = NumpyBackend()
            matrix = gate.asmatrix(backend)
            new.add(gate.__class__(matrix, *qubits, **gate.init_kwargs))
        else:
            new.add(gate.__class__(*qubits, **gate.init_kwargs))
        if len(qubits) == 2:
            add_swap = True

    return new, hardware_qubits


def respects_connectivity(circuit, middle_qubit=2, verbose=True):
    """Checks if a circuit respects connectivity constraints.

    Args:
        circuit (qibo.models.Circuit): Circuit model to check.
        middle_qubit (int): Hardware middle qubit.
        verbose (bool): If ``True`` it prints debugging log messages.

    Returns ``True`` if the following conditions are satisfied:
        - Circuit does not contain more than two-qubit gates.
        - All two-qubit gates have qubit 0 as target or control.
    otherwise returns ``False``.
    """
    # pring messages only if ``verbose == True``
    vlog = lambda msg: log.info(msg) if verbose else lambda msg: None
    for gate in circuit.queue:
        if len(gate.qubits) > 2 and not isinstance(gate, gates.M):
            vlog(f"{gate.name} acts on more than two qubits.")
            return False
        elif len(gate.qubits) == 2:
            if middle_qubit not in gate.qubits:
                vlog("Circuit does not respect connectivity. " f"{gate.name} acts on {gate.qubits}.")
                return False

    vlog("Circuit respects connectivity.")
    return True



from qibo.config import raise_error

from qibolab.platforms.abstract import AbstractPlatform


class DesignPlatform(AbstractPlatform):
    """Platform that using an instrument design.

    This will maybe replace the ``AbstractPlatform`` object
    and work as a generic platform that works with an arbitrary
    ``InstrumentDesign``.
    """

    def __init__(self, name, design, runcard):
        super().__init__(name, runcard)
        self.design = design

    def connect(self):
        self.design.connect()
        self.is_connected = True

    def setup(self):
        self.design.setup()

    def start(self):
        self.design.start()

    def stop(self):
        self.design.stop()

    def disconnect(self):
        self.design.disconnect()
        self.is_connected = False

    def execute_pulse_sequence(self, sequence, nshots=1024, relaxation_time=None):
        if relaxation_time is None:
            relaxation_time = self.relaxation_time
        return self.design.play(self.qubits, sequence, nshots=nshots, relaxation_time=relaxation_time)

    def sweep(self, sequence, *sweepers, nshots=1024, relaxation_time=None, average=True):
        if relaxation_time is None:
            relaxation_time = self.relaxation_time
        return self.design.sweep(
            self.qubits, sequence, *sweepers, nshots=nshots, relaxation_time=relaxation_time, average=average
        )

    def set_lo_drive_frequency(self, qubit, freq):
        self.qubits[qubit].drive.local_oscillator.frequency = freq

    def get_lo_drive_frequency(self, qubit):
        return self.qubits[qubit].drive.local_oscillator.frequency

    def set_lo_readout_frequency(self, qubit, freq):
        self.qubits[qubit].readout.local_oscillator.frequency = freq

    def get_lo_readout_frequency(self, qubit):
        return self.qubits[qubit].readout.local_oscillator.frequency

    def set_attenuation(self, qubit, att):
        raise_error(NotImplementedError, f"{self.name} does not support attenuation.")

    def get_attenuation(self, qubit):
        raise_error(NotImplementedError, f"{self.name} does not support attenuation.")

    def set_gain(self, qubit, gain):
        raise_error(NotImplementedError, f"{self.name} does not support gain.")

    def get_gain(self, qubit):
        raise_error(NotImplementedError, f"{self.name} does not support gain.")

    def set_bias(self, qubit, bias):
        self.qubits[qubit].flux.bias = bias

    def get_bias(self, qubit):
        return self.qubits[qubit].flux.bias



from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, List, Optional

import numpy as np
import yaml
from qibo import gates
from qibo.config import log, raise_error
from qibo.models import Circuit

from qibolab.designs.channels import Channel
from qibolab.pulses import FluxPulse, Pulse, PulseSequence, ReadoutPulse
from qibolab.transpilers import can_execute, transpile


@dataclass
class Qubit:
    """Representation of a physical qubit.

    Qubit objects are instantiated by :class:`qibolab.platforms.platform.Platform`
    but they are passed to instrument designs in order to play pulses.

    Args:
        name (int, str): Qubit number or name.
        readout (:class:`qibolab.platforms.utils.Channel`): Channel used to
            readout pulses to the qubit.
        feedback (:class:`qibolab.platforms.utils.Channel`): Channel used to
            get readout feedback from the qubit.
        drive (:class:`qibolab.platforms.utils.Channel`): Channel used to
            send drive pulses to the qubit.
        flux (:class:`qibolab.platforms.utils.Channel`): Channel used to
            send flux pulses to the qubit.
        Other characterization parameters for the qubit, loaded from the runcard.
    """

    name: str

    readout_frequency: int = 0
    drive_frequency: int = 0
    sweetspot: float = 0
    peak_voltage: float = 0
    pi_pulse_amplitude: float = 0
    T1: int = 0
    T2: int = 0
    state0_voltage: int = 0
    state1_voltage: int = 0
    mean_gnd_states: complex = 0 + 0.0j
    mean_exc_states: complex = 0 + 0.0j
    resonator_polycoef_flux: List[float] = field(default_factory=list)

    # filters used for applying CZ gate
    filter: dict = field(default_factory=dict)
    # parameters for single shot classification
    threshold: Optional[float] = None
    iq_angle: float = 0.0
    # required for integration weights (not sure if it should be here)
    rotation_angle: float = 0.0
    # required for mixers (not sure if it should be here)
    mixer_drive_g: float = 0.0
    mixer_drive_phi: float = 0.0
    mixer_readout_g: float = 0.0
    mixer_readout_phi: float = 0.0

    readout: Optional[Channel] = None
    feedback: Optional[Channel] = None
    twpa: Optional[Channel] = None
    drive: Optional[Channel] = None
    flux: Optional[Channel] = None

    def __post_init__(self):
        # register qubit in ``flux`` channel so that we can access
        # ``sweetspot`` and ``filters`` at the channel level
        if self.flux:
            self.flux.qubit = self

    @property
    def channels(self):
        for channel in [self.readout, self.feedback, self.drive, self.flux, self.twpa]:
            if channel is not None:
                yield channel


class AbstractPlatform(ABC):
    """Abstract platform for controlling quantum devices.

    Args:
        name (str): name of the platform.
        runcard (str): path to the yaml file containing the platform setup.
    """

    def __init__(self, name, runcard):
        log.info(f"Loading platform {name} from runcard {runcard}")
        self.name = name
        self.runcard = runcard

        self.qubits = {}

        # Values for the following are set from the runcard in ``reload_settings``
        self.settings = None
        self.is_connected = False
        self.nqubits = None
        self.resonator_type = None
        self.topology = None
        self.relaxation_time = None
        self.sampling_rate = None

        self.native_single_qubit_gates = {}
        self.native_two_qubit_gates = {}
        self.two_qubit_natives = set()
        # Load platform settings
        self.reload_settings()

    def __repr__(self):
        return self.name

    def _check_connected(self):
        if not self.is_connected:  # pragma: no cover
            raise_error(RuntimeError, "Cannot access instrument because it is not connected.")

    def reload_settings(self):
        # TODO: Remove ``self.settings``
        with open(self.runcard) as file:
            settings = self.settings = yaml.safe_load(file)

        self.nqubits = settings["nqubits"]
        if "resonator_type" in self.settings:
            self.resonator_type = self.settings["resonator_type"]
        else:
            self.resonator_type = "3D" if self.nqubits == 1 else "2D"

        self.topology = settings["topology"]

        self.relaxation_time = settings["settings"]["repetition_duration"]
        self.sampling_rate = settings["settings"]["sampling_rate"]

        # TODO: Create better data structures for native gates
        self.native_gates = settings["native_gates"]
        self.native_single_qubit_gates = self.native_gates["single_qubit"]
        if "two_qubit" in self.native_gates:
            self.native_two_qubit_gates = self.native_gates["two_qubit"]
            for gates in self.native_gates["two_qubit"].values():
                self.two_qubit_natives |= set(gates.keys())
        else:
            # dummy value to avoid transpiler failure for single qubit devices
            self.two_qubit_natives = {"CZ"}

        # Load characterization settings and create ``Qubit`` and ``Channel`` objects
        for q in settings["qubits"]:
            if q in self.qubits:
                for name, value in settings["characterization"]["single_qubit"][q].items():
                    setattr(self.qubits[q], name, value)
            else:
                self.qubits[q] = Qubit(q, **settings["characterization"]["single_qubit"][q])

    @abstractmethod
    def connect(self):
        """Connects to instruments."""

    @abstractmethod
    def setup(self):
        """Prepares instruments to execute experiments."""

    @abstractmethod
    def start(self):
        """Starts all the instruments."""

    @abstractmethod
    def stop(self):
        """Starts all the instruments."""

    @abstractmethod
    def disconnect(self):
        """Disconnects to instruments."""

    def transpile(self, circuit: Circuit):
        """Transforms a circuit to pulse sequence.

        Args:
            circuit (qibo.models.Circuit): Qibo circuit that respects the platform's
                connectivity and native gates.

        Returns:
            sequence (qibolab.pulses.PulseSequence): Pulse sequence that implements the
                circuit on the qubit.
        """
        if not can_execute(circuit, self.two_qubit_natives):
            circuit, _ = transpile(circuit, self.two_qubit_natives)

        sequence = PulseSequence()
        virtual_z_phases = defaultdict(int)

        # keep track of gates that were already added to avoid adding them twice
        already_processed = set()
        # process circuit gates
        for moment in circuit.queue.moments:
            moment_start = sequence.finish
            for gate in moment:
                if isinstance(gate, gates.I) or gate is None or gate in already_processed:
                    pass

                elif isinstance(gate, gates.Z):
                    qubit = gate.target_qubits[0]
                    virtual_z_phases[qubit] += np.pi

                elif isinstance(gate, gates.RZ):
                    qubit = gate.target_qubits[0]
                    virtual_z_phases[qubit] += gate.parameters[0]

                elif isinstance(gate, gates.U3):
                    qubit = gate.target_qubits[0]
                    # Transform gate to U3 and add pi/2-pulses
                    theta, phi, lam = gate.parameters
                    # apply RZ(lam)
                    virtual_z_phases[qubit] += lam
                    # Fetch pi/2 pulse from calibration
                    RX90_pulse_1 = self.create_RX90_pulse(
                        qubit,
                        start=max(sequence.get_qubit_pulses(qubit).finish, moment_start),
                        relative_phase=virtual_z_phases[qubit],
                    )
                    # apply RX(pi/2)
                    sequence.add(RX90_pulse_1)
                    # apply RZ(theta)
                    virtual_z_phases[qubit] += theta
                    # Fetch pi/2 pulse from calibration
                    RX90_pulse_2 = self.create_RX90_pulse(
                        qubit, start=RX90_pulse_1.finish, relative_phase=virtual_z_phases[qubit] - np.pi
                    )
                    # apply RX(-pi/2)
                    sequence.add(RX90_pulse_2)
                    # apply RZ(phi)
                    virtual_z_phases[qubit] += phi

                elif isinstance(gate, gates.M):
                    # Add measurement pulse
                    measurement_start = max(sequence.get_qubit_pulses(*gate.target_qubits).finish, moment_start)
                    gate.pulses = ()
                    for qubit in gate.target_qubits:
                        MZ_pulse = self.create_MZ_pulse(qubit, start=measurement_start)
                        sequence.add(MZ_pulse)
                        gate.pulses = (*gate.pulses, MZ_pulse.serial)

                elif isinstance(gate, gates.CZ):
                    # create CZ pulse sequence with start time = 0
                    cz_sequence, cz_virtual_z_phases = self.create_CZ_pulse_sequence(gate.qubits)

                    # determine the right start time based on the availability of the qubits involved
                    cz_qubits = {*cz_sequence.qubits, *gate.qubits}
                    cz_start = max(sequence.get_qubit_pulses(*cz_qubits).finish, moment_start)

                    # shift the pulses
                    for pulse in cz_sequence.pulses:
                        pulse.start += cz_start

                    # add pulses to the sequence
                    sequence.add(cz_sequence)

                    # update z_phases registers
                    for qubit in cz_virtual_z_phases:
                        virtual_z_phases[qubit] += cz_virtual_z_phases[qubit]

                else:  # pragma: no cover
                    raise_error(
                        NotImplementedError,
                        f"Transpilation of {gate.__class__.__name__} gate has not been implemented yet.",
                    )

                already_processed.add(gate)
        return sequence

    @abstractmethod
    def execute_pulse_sequence(self, sequence, nshots=1024, relaxation_time=None):
        """Executes a pulse sequence.

        Args:
            sequence (:class:`qibolab.pulses.PulseSequence`): Pulse sequence to execute.
            nshots (int): Number of shots to sample from the experiment. Default is 1024.
            relaxation_time (int): Time to wait for the qubit to relax to its ground state between shots in ns.
                If ``None`` the default value provided as ``repetition_duration`` in the runcard will be used.

        Returns:
            Readout results acquired by after execution.
        """

    def __call__(self, sequence, nshots=1024, relaxation_time=None):
        return self.execute_pulse_sequence(sequence, nshots, relaxation_time)

    def sweep(self, sequence, *sweepers, nshots=1024, average=True, relaxation_time=None):
        """Executes a pulse sequence for different values of sweeped parameters.
        Useful for performing chip characterization.

        Example:
            .. testcode::

                import numpy as np
                from qibolab.platform import Platform
                from qibolab.sweeper import Sweeper, Parameter
                from qibolab.pulses import PulseSequence


                platform = Platform("dummy")
                sequence = PulseSequence()
                parameter = Parameter.frequency
                pulse = platform.create_qubit_readout_pulse(qubit=0, start=0)
                sequence.add(pulse)
                parameter_range = np.random.randint(10, size=10)
                sweeper = Sweeper(parameter, parameter_range, [pulse])
                platform.sweep(sequence, sweeper)


        Args:
            sequence (:class:`qibolab.pulses.PulseSequence`): Pulse sequence to execute.
            sweepers (:class:`qibolab.sweeper.Sweeper`): Sweeper objects that specify which
                parameters are being sweeped.
            nshots (int): Number of shots to sample from the experiment. Default is 1024.
            relaxation_time (int): Time to wait for the qubit to relax to its ground state between shots in ns.
                If ``None`` the default value provided as ``repetition_duration`` in the runcard will be used.
            average (bool): If ``True`` the IQ results of individual shots are averaged on hardware.

        Returns:
            Readout results acquired by after execution.
        """
        raise_error(NotImplementedError, f"Platform {self.name} does not support sweeping.")

    def get_qd_channel(self, qubit):
        if self.qubits[qubit].drive:
            return self.qubits[qubit].drive.name
        else:
            return self.settings["qubit_channel_map"][qubit][1]

    # TODO: Maybe create a dataclass for native gates
    def create_RX90_pulse(self, qubit, start=0, relative_phase=0):
        pulse_kwargs = self.native_single_qubit_gates[qubit]["RX"]
        qd_duration = pulse_kwargs["duration"]
        qd_frequency = pulse_kwargs["frequency"]
        qd_amplitude = pulse_kwargs["amplitude"] / 2.0
        qd_shape = pulse_kwargs["shape"]
        qd_channel = self.get_qd_channel(qubit)
        return Pulse(start, qd_duration, qd_amplitude, qd_frequency, relative_phase, qd_shape, qd_channel, qubit=qubit)

    def create_RX_pulse(self, qubit, start=0, relative_phase=0):
        pulse_kwargs = self.native_single_qubit_gates[qubit]["RX"]
        qd_duration = pulse_kwargs["duration"]
        qd_frequency = pulse_kwargs["frequency"]
        qd_amplitude = pulse_kwargs["amplitude"]
        qd_shape = pulse_kwargs["shape"]
        qd_channel = self.get_qd_channel(qubit)
        return Pulse(start, qd_duration, qd_amplitude, qd_frequency, relative_phase, qd_shape, qd_channel, qubit=qubit)

    def create_CZ_pulse_sequence(self, qubits, start=0):
        # Check in the settings if qubits[0]-qubits[1] is a key
        if f"{qubits[0]}-{qubits[1]}" in self.settings["native_gates"]["two_qubit"]:
            pulse_sequence_settings = self.settings["native_gates"]["two_qubit"][f"{qubits[0]}-{qubits[1]}"]["CZ"]
        elif f"{qubits[1]}-{qubits[0]}" in self.settings["native_gates"]["two_qubit"]:
            pulse_sequence_settings = self.settings["native_gates"]["two_qubit"][f"{qubits[1]}-{qubits[0]}"]["CZ"]
        else:
            raise_error(
                ValueError,
                f"Calibration for CZ gate between qubits {qubits[0]} and {qubits[1]} not found.",
            )

        # If settings contains only one pulse dictionary, convert it into a list that can be iterated below
        if isinstance(pulse_sequence_settings, dict):
            pulse_sequence_settings = [pulse_sequence_settings]

        from qibolab.pulses import FluxPulse, PulseSequence

        sequence = PulseSequence()
        virtual_z_phases = defaultdict(int)

        for pulse_settings in pulse_sequence_settings:
            if pulse_settings["type"] == "qf":
                qf_duration = pulse_settings["duration"]
                qf_amplitude = pulse_settings["amplitude"]
                qf_shape = pulse_settings["shape"]
                qubit = pulse_settings["qubit"]
                if self.qubits[qubit].flux:
                    qf_channel = self.qubits[qubit].flux.name
                else:
                    qf_channel = self.settings["qubit_channel_map"][qubit][2]
                sequence.add(
                    FluxPulse(
                        start + pulse_settings["relative_start"], qf_duration, qf_amplitude, qf_shape, qf_channel, qubit
                    )
                )
            elif pulse_settings["type"] == "virtual_z":
                virtual_z_phases[pulse_settings["qubit"]] += pulse_settings["phase"]
            else:
                raise NotImplementedError(
                    "Implementation of CZ gates using pulses of types other than `qf` or `virtual_z` is not supported yet."
                )

        return sequence, virtual_z_phases

    def create_MZ_pulse(self, qubit, start):
        pulse_kwargs = self.native_single_qubit_gates[qubit]["MZ"]
        ro_duration = pulse_kwargs["duration"]
        ro_frequency = pulse_kwargs["frequency"]
        ro_amplitude = pulse_kwargs["amplitude"]
        ro_shape = pulse_kwargs["shape"]
        if self.qubits[qubit].readout:
            ro_channel = self.qubits[qubit].readout.name
        else:
            ro_channel = self.settings["qubit_channel_map"][qubit][0]
        return ReadoutPulse(start, ro_duration, ro_amplitude, ro_frequency, 0, ro_shape, ro_channel, qubit=qubit)

    def create_qubit_drive_pulse(self, qubit, start, duration, relative_phase=0):
        pulse_kwargs = self.native_single_qubit_gates[qubit]["RX"]
        qd_frequency = pulse_kwargs["frequency"]
        qd_amplitude = pulse_kwargs["amplitude"]
        qd_shape = pulse_kwargs["shape"]
        qd_channel = self.get_qd_channel(qubit)
        return Pulse(start, duration, qd_amplitude, qd_frequency, relative_phase, qd_shape, qd_channel, qubit=qubit)

    def create_qubit_readout_pulse(self, qubit, start):
        return self.create_MZ_pulse(qubit, start)

    # TODO Remove RX90_drag_pulse and RX_drag_pulse, replace them with create_qubit_drive_pulse
    # TODO Add RY90 and RY pulses

    def create_RX90_drag_pulse(self, qubit, start, relative_phase=0, beta=None):
        # create RX pi/2 pulse with drag shape
        pulse_kwargs = self.native_single_qubit_gates[qubit]["RX"]
        qd_duration = pulse_kwargs["duration"]
        qd_frequency = pulse_kwargs["frequency"]
        qd_amplitude = pulse_kwargs["amplitude"] / 2.0
        qd_shape = pulse_kwargs["shape"]
        if beta != None:
            qd_shape = "Drag(5," + str(beta) + ")"

        qd_channel = self.get_qd_channel(qubit)
        return Pulse(start, qd_duration, qd_amplitude, qd_frequency, relative_phase, qd_shape, qd_channel, qubit=qubit)

    def create_RX_drag_pulse(self, qubit, start, relative_phase=0, beta=None):
        # create RX pi pulse with drag shape
        pulse_kwargs = self.native_single_qubit_gates[qubit]["RX"]
        qd_duration = pulse_kwargs["duration"]
        qd_frequency = pulse_kwargs["frequency"]
        qd_amplitude = pulse_kwargs["amplitude"]
        qd_shape = pulse_kwargs["shape"]
        if beta != None:
            qd_shape = "Drag(5," + str(beta) + ")"

        qd_channel = self.get_qd_channel(qubit)
        return Pulse(start, qd_duration, qd_amplitude, qd_frequency, relative_phase, qd_shape, qd_channel, qubit=qubit)

    @abstractmethod
    def set_lo_drive_frequency(self, qubit, freq):
        """Set frequency of the qubit drive local oscillator.

        Args:
            qubit (int): qubit whose local oscillator will be modified.
            freq (int): new value of the frequency in Hz.
        """

    @abstractmethod
    def get_lo_drive_frequency(self, qubit):
        """Get frequency of the qubit drive local oscillator in Hz."""

    @abstractmethod
    def set_lo_readout_frequency(self, qubit, freq):
        """Set frequency of the qubit drive local oscillator.

        Args:
            qubit (int): qubit whose local oscillator will be modified.
            freq (int): new value of the frequency in Hz.
        """

    @abstractmethod
    def get_lo_readout_frequency(self, qubit):
        """Get frequency of the qubit readout local oscillator in Hz."""

    @abstractmethod
    def set_attenuation(self, qubit, att):
        """Set attenuation value. Usefeul for calibration routines such as punchout.

        Args:
            qubit (int): qubit whose attenuation will be modified.
            att (int): new value of the attenuation (dB).
        Returns:
            None
        """

    @abstractmethod
    def get_attenuation(self, qubit):
        """Get attenuation value. Usefeul for calibration routines such as punchout."""

    @abstractmethod
    def set_gain(self, qubit, gain):
        """Set gain value. Usefeul for calibration routines such as Rabi oscillations.

        Args:
            qubit (int): qubit whose attenuation will be modified.
            gain (int): new value of the gain (dimensionless).
        Returns:
            None
        """

    @abstractmethod
    def get_gain(self, qubit):
        """Get gain value. Usefeul for calibration routines such as Rabi oscillations."""

    @abstractmethod
    def set_bias(self, qubit, bias):
        """Set bias value. Usefeul for calibration routines involving flux.

        Args:
            qubit (int): qubit whose attenuation will be modified.
            bias (int): new value of the bias (V).
        Returns:
            None
        """

    @abstractmethod
    def get_bias(self, qubit):
        """Get bias value. Usefeul for calibration routines involving flux."""



import collections
from dataclasses import dataclass, field

import numpy as np
from qibo.config import log, raise_error
from qm.qua import (
    align,
    assign,
    declare,
    declare_stream,
    dual_demod,
    fixed,
    for_,
    frame_rotation_2pi,
    measure,
    play,
    program,
    reset_frame,
    reset_phase,
    save,
    stream_processing,
    wait,
)
from qm.QuantumMachinesManager import QuantumMachinesManager
from qualang_tools.bakery import baking
from qualang_tools.loops import from_array

from qibolab.instruments.abstract import AbstractInstrument
from qibolab.pulses import Pulse, PulseType, Rectangular
from qibolab.result import ExecutionResults
from qibolab.sweeper import Parameter


@dataclass
class QMConfig:
    """Configuration for communicating with the ``QuantumMachinesManager``."""

    version: int = 1
    controllers: dict = field(default_factory=dict)
    elements: dict = field(default_factory=dict)
    pulses: dict = field(default_factory=dict)
    waveforms: dict = field(default_factory=dict)
    digital_waveforms: dict = field(default_factory=lambda: {"ON": {"samples": [(1, 0)]}})
    integration_weights: dict = field(default_factory=dict)
    mixers: dict = field(default_factory=dict)

    def register_analog_output_controllers(self, ports, offset=0.0, filter=None):
        """Register controllers in the ``config``.

        Args:
            ports (list): List of tuples ``(conX, port)``.
            offset (float): Constant offset to be played in the given ports.
                Relevant for ports connected to flux channels.
            filter (dict): Pulse shape filters. Relevant for ports connected to flux channels.
                QM syntax should be followed for the filters.
        """
        if abs(offset) > 0.2:
            raise_error(ValueError, f"DC offset for Quantum Machines cannot exceed 0.1V but is {offset}.")

        for con, port in ports:
            if con not in self.controllers:
                self.controllers[con] = {"analog_outputs": {}}
            self.controllers[con]["analog_outputs"][port] = {"offset": offset}
            if filter is not None:
                self.controllers[con]["analog_outputs"][port]["filter"] = filter

    @staticmethod
    def iq_imbalance(g, phi):
        """Creates the correction matrix for the mixer imbalance caused by the gain and phase imbalances

        More information here:
        https://docs.qualang.io/libs/examples/mixer-calibration/#non-ideal-mixer

        Args:
            g (float): relative gain imbalance between the I & Q ports (unit-less).
                Set to 0 for no gain imbalance.
            phi (float): relative phase imbalance between the I & Q ports (radians).
                Set to 0 for no phase imbalance.
        """
        c = np.cos(phi)
        s = np.sin(phi)
        N = 1 / ((1 - g**2) * (2 * c**2 - 1))
        return [float(N * x) for x in [(1 - g) * c, (1 + g) * s, (1 - g) * s, (1 + g) * c]]

    def register_drive_element(self, qubit, intermediate_frequency=0):
        """Register qubit drive elements and controllers in the QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit to add elements for.
            intermediate_frequency (int): Intermediate frequency that the OPX
                will send to this qubit. This frequency will be mixed with the
                LO connected to the same channel.
        """
        if f"drive{qubit.name}" not in self.elements:
            # register drive controllers
            self.register_analog_output_controllers(qubit.drive.ports)
            # register element
            lo_frequency = int(qubit.drive.local_oscillator.frequency)
            self.elements[f"drive{qubit.name}"] = {
                "mixInputs": {
                    "I": qubit.drive.ports[0],
                    "Q": qubit.drive.ports[1],
                    "lo_frequency": lo_frequency,
                    "mixer": f"mixer_drive{qubit.name}",
                },
                "intermediate_frequency": intermediate_frequency,
                "operations": {},
            }
            drive_g = qubit.mixer_drive_g
            drive_phi = qubit.mixer_drive_phi
            self.mixers[f"mixer_drive{qubit.name}"] = [
                {
                    "intermediate_frequency": intermediate_frequency,
                    "lo_frequency": lo_frequency,
                    "correction": self.iq_imbalance(drive_g, drive_phi),
                }
            ]
        else:
            self.elements[f"drive{qubit.name}"]["intermediate_frequency"] = intermediate_frequency
            self.mixers[f"mixer_drive{qubit.name}"][0]["intermediate_frequency"] = intermediate_frequency

    def register_readout_element(self, qubit, intermediate_frequency=0, time_of_flight=0, smearing=0):
        """Register resonator elements and controllers in the QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit to add elements for.
            intermediate_frequency (int): Intermediate frequency that the OPX
                will send to this qubit. This frequency will be mixed with the
                LO connected to the same channel.
        """
        if f"readout{qubit.name}" not in self.elements:
            # register readout controllers
            self.register_analog_output_controllers(qubit.readout.ports)
            # register feedback controllers
            controllers = self.controllers
            for con, port in qubit.feedback.ports:
                if con not in controllers:
                    controllers[con] = {
                        "analog_outputs": {},
                        "digital_outputs": {
                            1: {},
                        },
                        "analog_inputs": {},
                    }
                if "digital_outputs" not in controllers[con]:
                    controllers[con]["digital_outputs"] = {
                        1: {},
                    }
                if "analog_inputs" not in controllers[con]:
                    controllers[con]["analog_inputs"] = {}
                controllers[con]["analog_inputs"][port] = {"offset": 0.0, "gain_db": 0}

            # register element
            lo_frequency = int(qubit.readout.local_oscillator.frequency)
            self.elements[f"readout{qubit.name}"] = {
                "mixInputs": {
                    "I": qubit.readout.ports[0],
                    "Q": qubit.readout.ports[1],
                    "lo_frequency": lo_frequency,
                    "mixer": f"mixer_readout{qubit.name}",
                },
                "intermediate_frequency": intermediate_frequency,
                "operations": {},
                "outputs": {
                    "out1": qubit.feedback.ports[0],
                    "out2": qubit.feedback.ports[1],
                },
                "time_of_flight": time_of_flight,
                "smearing": smearing,
            }
            readout_g = qubit.mixer_readout_g
            readout_phi = qubit.mixer_readout_phi
            self.mixers[f"mixer_readout{qubit.name}"] = [
                {
                    "intermediate_frequency": intermediate_frequency,
                    "lo_frequency": lo_frequency,
                    "correction": self.iq_imbalance(readout_g, readout_phi),
                }
            ]
        else:
            self.elements[f"readout{qubit.name}"]["intermediate_frequency"] = intermediate_frequency
            self.mixers[f"mixer_readout{qubit.name}"][0]["intermediate_frequency"] = intermediate_frequency

    def register_flux_element(self, qubit, intermediate_frequency=0):
        """Register qubit flux elements and controllers in the QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit to add elements for.
            intermediate_frequency (int): Intermediate frequency that the OPX
                will send to this qubit. This frequency will be mixed with the
                LO connected to the same channel.
        """
        if f"flux{qubit.name}" not in self.elements:
            # register controller
            self.register_analog_output_controllers(qubit.flux.ports, qubit.flux.bias, qubit.flux.filter)
            # register element
            self.elements[f"flux{qubit.name}"] = {
                "singleInput": {
                    "port": qubit.flux.ports[0],
                },
                "intermediate_frequency": intermediate_frequency,
                "operations": {},
            }
        else:
            self.elements[f"flux{qubit.name}"]["intermediate_frequency"] = intermediate_frequency

    def register_pulse(self, qubit, pulse, time_of_flight, smearing):
        """Registers pulse, waveforms and integration weights in QM config.

        Args:
            qubit (:class:`qibolab.platforms.utils.Qubit`): Qubit that the pulse acts on.
            pulse (:class:`qibolab.pulses.Pulse`): Pulse object to register.

        Returns:
            element (str): Name of the element this pulse will be played on.
                Elements are a part of the QM config and are generated during
                instantiation of the Qubit objects. They are named as
                "drive0", "drive1", "flux0", "readout0", ...
        """
        if pulse.serial not in self.pulses:
            if pulse.type.name == "DRIVE":
                serial_i = self.register_waveform(pulse, "i")
                serial_q = self.register_waveform(pulse, "q")
                self.pulses[pulse.serial] = {
                    "operation": "control",
                    "length": pulse.duration,
                    "waveforms": {"I": serial_i, "Q": serial_q},
                }
                # register drive element (if it does not already exist)
                if_frequency = pulse.frequency - int(qubit.drive.local_oscillator.frequency)
                self.register_drive_element(qubit, if_frequency)
                # register flux element (if available)
                if qubit.flux:
                    self.register_flux_element(qubit)
                # register drive pulse in elements
                self.elements[f"drive{qubit.name}"]["operations"][pulse.serial] = pulse.serial

            elif pulse.type.name == "FLUX":
                serial = self.register_waveform(pulse)
                self.pulses[pulse.serial] = {
                    "operation": "control",
                    "length": pulse.duration,
                    "waveforms": {
                        "single": serial,
                    },
                }
                # register flux element (if it does not already exist)
                self.register_flux_element(qubit, pulse.frequency)
                # register flux pulse in elements
                self.elements[f"flux{qubit.name}"]["operations"][pulse.serial] = pulse.serial

            elif pulse.type.name == "READOUT":
                serial_i = self.register_waveform(pulse, "i")
                serial_q = self.register_waveform(pulse, "q")
                self.register_integration_weights(qubit, pulse.duration)
                self.pulses[pulse.serial] = {
                    "operation": "measurement",
                    "length": pulse.duration,
                    "waveforms": {
                        "I": serial_i,
                        "Q": serial_q,
                    },
                    "integration_weights": {
                        "cos": f"cosine_weights{qubit.name}",
                        "sin": f"sine_weights{qubit.name}",
                        "minus_sin": f"minus_sine_weights{qubit.name}",
                    },
                    "digital_marker": "ON",
                }
                # register readout element (if it does not already exist)
                if_frequency = pulse.frequency - int(qubit.readout.local_oscillator.frequency)
                self.register_readout_element(qubit, if_frequency, time_of_flight, smearing)
                # register flux element (if available)
                if qubit.flux:
                    self.register_flux_element(qubit)
                # register readout pulse in elements
                self.elements[f"readout{qubit.name}"]["operations"][pulse.serial] = pulse.serial

            else:
                raise_error(TypeError, f"Unknown pulse type {pulse.type.name}.")

    def register_waveform(self, pulse, mode="i"):
        """Registers waveforms in QM config.

        QM supports two kinds of waveforms, examples:
            "zero_wf": {"type": "constant", "sample": 0.0}
            "x90_wf": {"type": "arbitrary", "samples": x90_wf.tolist()}

        Args:
            pulse (:class:`qibolab.pulses.Pulse`): Pulse object to read the waveform from.
            mode (str): "i" or "q" specifying which channel the waveform will be played.

        Returns:
            serial (str): String with a serialization of the waveform.
                Used as key to identify the waveform in the config.
        """
        # Maybe need to force zero q waveforms
        # if pulse.type.name == "READOUT" and mode == "q":
        #    serial = "zero_wf"
        #    if serial not in self.waveforms:
        #        self.waveforms[serial] = {"type": "constant", "sample": 0.0}
        if isinstance(pulse.shape, Rectangular):
            serial = f"constant_wf{pulse.amplitude}"
            if serial not in self.waveforms:
                self.waveforms[serial] = {"type": "constant", "sample": pulse.amplitude}
        else:
            waveform = getattr(pulse, f"envelope_waveform_{mode}")
            serial = waveform.serial
            if serial not in self.waveforms:
                self.waveforms[serial] = {"type": "arbitrary", "samples": waveform.data.tolist()}
        return serial

    def register_integration_weights(self, qubit, readout_len):
        """Registers integration weights in QM config.

        Args:
            qubit (:class:`qibolab.platforms.quantum_machines.Qubit`): Qubit
                object that the integration weights will be used for.
            readout_len (int): Duration of the readout pulse in ns.
        """
        rotation_angle = qubit.rotation_angle
        self.integration_weights.update(
            {
                f"cosine_weights{qubit.name}": {
                    "cosine": [(np.cos(rotation_angle), readout_len)],
                    "sine": [(-np.sin(rotation_angle), readout_len)],
                },
                f"sine_weights{qubit.name}": {
                    "cosine": [(np.sin(rotation_angle), readout_len)],
                    "sine": [(np.cos(rotation_angle), readout_len)],
                },
                f"minus_sine_weights{qubit.name}": {
                    "cosine": [(-np.sin(rotation_angle), readout_len)],
                    "sine": [(-np.cos(rotation_angle), readout_len)],
                },
            }
        )


class QMPulse:
    def __init__(self, pulse):
        self.pulse = pulse
        self.element = f"{pulse.type.name.lower()}{pulse.qubit}"
        self.operation = pulse.serial
        self.relative_phase = pulse.relative_phase / (2 * np.pi)
        self.duration = pulse.duration

        # Stores the baking object (for pulses that need 1ns resolution)
        self.baked = None
        self.baked_amplitude = None

        self.I = None
        self.Q = None
        self.shot = None
        self.I_st = None
        self.Q_st = None
        self.shots = None
        self.threshold = None
        self.cos = None
        self.sin = None

    def declare_output(self, threshold=None, iq_angle=None):
        self.I = declare(fixed)
        self.Q = declare(fixed)
        self.I_st = declare_stream()
        self.Q_st = declare_stream()

        if threshold is not None:
            # QUA variables used for single shot classification
            self.shot = declare(bool)
            self.shots = declare_stream()
            self.threshold = threshold
            self.cos = np.cos(iq_angle)
            self.sin = np.sin(iq_angle)

    def bake(self, config: QMConfig):
        if self.baked is not None:
            raise_error(RuntimeError, f"Bake was already called for {self.pulse}.")
        # ! Only works for flux pulses that have zero Q waveform

        # Create the different baked sequences, each one corresponding to a different truncated duration
        # for t in range(self.pulse.duration + 1):
        with baking(config.__dict__, padding_method="right") as self.baked:
            # if t == 0:  # Otherwise, the baking will be empty and will not be created
            #    wf = [0.0] * 16
            # else:
            # wf = waveform[:t].tolist()
            if self.pulse.duration == 0:
                waveform = [0.0] * 16
            else:
                waveform = self.pulse.envelope_waveform_i.data.tolist()
            self.baked.add_op(self.pulse.serial, self.element, waveform)
            self.baked.play(self.pulse.serial, self.element)
            # Append the baking object in the list to call it from the QUA program
            # self.segments.append(b)

        self.duration = self.baked.get_op_length()


class QMSequence(list):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # keep track of readout pulses for registering outputs
        self.ro_pulses = []
        # map from qibolab pulses to QMPulses (useful when sweeping)
        self.pulse_to_qmpulse = {}

    def add(self, pulse):
        if not isinstance(pulse, Pulse):
            raise_error(TypeError, f"Pulse {pulse} has invalid type {type(pulse)}.")

        qmpulse = QMPulse(pulse)
        self.pulse_to_qmpulse[pulse.serial] = qmpulse
        if pulse.type.name == "READOUT":
            self.ro_pulses.append(qmpulse)
        super().append(qmpulse)
        return qmpulse


class QMOPX(AbstractInstrument):
    """Instrument object for controlling Quantum Machines (QM) OPX controllers.

    Playing pulses on QM controllers requires a ``config`` dictionary and a program
    written in QUA language. The ``config`` file is generated in parts in the following places
    in the ``register_*`` methods. The controllers, elements and pulses are all
    registered after a pulse sequence is given, so that the config contains only
    elements related to the participating qubits.
    The QUA program for executing an arbitrary qibolab ``PulseSequence`` is written in
    ``play`` and ``play_pulses`` and executed in ``execute_program``.

    Args:
        name (str): Name of the instrument instance.
        address (str): IP address and port for connecting to the OPX instruments.

    Attributes:
        is_connected (bool): Boolean that shows whether instruments are connected.
        manager (:class:`qm.QuantumMachinesManager.QuantumMachinesManager`): Manager object
            used for controlling the QM OPXs.
        config (dict): Configuration dictionary required for pulse execution on the OPXs.
        time_of_flight (int): Time of flight used for hardware signal integration.
        smearing (int): Smearing used for hardware signal integration.
    """

    def __init__(self, name, address):
        # QuantumMachines manager is instantiated in ``platform.connect``
        self.name = name
        self.address = address
        self.manager = None
        self.is_connected = False

        self.time_of_flight = 0
        self.smearing = 0
        # copied from qblox runcard, not used here yet
        # hardware_avg: 1024
        # sampling_rate: 1_000_000_000
        # repetition_duration: 200_000
        # minimum_delay_between_instructions: 4
        self.config = QMConfig()

    def connect(self):
        """Connect to the QM manager."""
        host, port = self.address.split(":")
        self.manager = QuantumMachinesManager(host, int(port))

    def setup(self):
        """Deprecated method."""
        # controllers are defined when registering pulses
        pass

    def start(self):
        # TODO: Start the OPX flux offsets?
        pass

    def stop(self):
        """Close all running Quantum Machines."""
        # TODO: Use logging
        # log.warn("Closing all Quantum Machines.")
        print("Closing all Quantum Machines.")
        self.manager.close_all_quantum_machines()

    def disconnect(self):
        """Disconnect from QM manager."""
        if self.is_connected:
            self.manager.close()
            self.is_connected = False

    def execute_program(self, program):
        """Executes an arbitrary program written in QUA language.

        Args:
            program: QUA program.

        Returns:
            TODO
        """
        machine = self.manager.open_qm(self.config.__dict__)

        # for debugging only
        from qm import generate_qua_script

        with open("qua_script.txt", "w") as file:
            file.write(generate_qua_script(program, self.config.__dict__))

        return machine.execute(program)

    @staticmethod
    def play_pulses(qmsequence, relaxation_time=0):
        """Part of QUA program that plays an arbitrary pulse sequence.

        Should be used inside a ``program()`` context.

        Args:
            qmsequence (list): Pulse sequence containing QM specific pulses (``qmpulse``).
                These pulses are defined in :meth:`qibolab.instruments.qm.QMOPX.play` and
                hold attributes for the ``element`` and ``operation`` that corresponds to
                each pulse, as defined in the QM config.
        """
        needs_reset = False
        align()
        clock = collections.defaultdict(int)
        for qmpulse in qmsequence:
            pulse = qmpulse.pulse
            wait_time = pulse.start - clock[qmpulse.element]
            if wait_time >= 12:
                wait_cycles = wait_time // 4 + 1
                wait(wait_cycles, qmpulse.element)
                clock[qmpulse.element] += 4 * wait_cycles
            clock[qmpulse.element] += qmpulse.duration
            if pulse.type.name == "READOUT":
                measure(
                    qmpulse.operation,
                    qmpulse.element,
                    None,
                    dual_demod.full("cos", "out1", "sin", "out2", qmpulse.I),
                    dual_demod.full("minus_sin", "out1", "cos", "out2", qmpulse.Q),
                )
                if qmpulse.threshold is not None:
                    assign(qmpulse.shot, qmpulse.I * qmpulse.cos - qmpulse.Q * qmpulse.sin > qmpulse.threshold)
            else:
                if not isinstance(qmpulse.relative_phase, float) or qmpulse.relative_phase != 0:
                    frame_rotation_2pi(qmpulse.relative_phase, qmpulse.element)
                    needs_reset = True
                if qmpulse.baked is not None:
                    if qmpulse.baked_amplitude is not None:
                        qmpulse.baked.run(amp_array=[(qmpulse.element, qmpulse.baked_amplitude)])
                    else:
                        qmpulse.baked.run()
                else:
                    play(qmpulse.operation, qmpulse.element)
                if needs_reset:
                    reset_frame(qmpulse.element)

        # for Rabi-length?
        if relaxation_time > 0:
            wait(relaxation_time // 4, *clock.keys())

        # Save data to the stream processing
        for qmpulse in qmsequence.ro_pulses:
            save(qmpulse.I, qmpulse.I_st)
            save(qmpulse.Q, qmpulse.Q_st)
            if qmpulse.threshold is not None:
                save(qmpulse.shot, qmpulse.shots)

    @staticmethod
    def fetch_results(result, ro_pulses):
        """Fetches results from an executed experiment."""
        # TODO: Update result asynchronously instead of waiting
        # for all values, in order to allow live plotting
        # import time
        # for _ in range(5):
        #    handles.is_processing()
        #    time.sleep(1)
        handles = result.result_handles
        handles.wait_for_all_values()
        results = {}
        for pulse in ro_pulses:
            serial = pulse.serial
            ires = handles.get(f"{serial}_I").fetch_all()
            qres = handles.get(f"{serial}_Q").fetch_all()
            if f"{serial}_shots" in handles:
                shots = handles.get(f"{serial}_shots").fetch_all().astype(int)
            else:
                shots = None
            results[pulse.qubit] = results[serial] = ExecutionResults.from_components(ires, qres, shots)
        return results

    def play(self, qubits, sequence, nshots, relaxation_time):
        """Plays an arbitrary pulse sequence using QUA program.

        Args:
            qubits (list): List of :class:`qibolab.platforms.utils.Qubit` objects
                passed from the platform.
            sequence (:class:`qibolab.pulses.PulseSequence`). Pulse sequence to play.
            nshots (int): Number of repetitions (shots) of the experiment.
            relaxation_time (int): Time to wait for the qubit to relax to its ground state between shots in ns.
        """
        # register flux elements for all qubits so that they are
        # always at sweetspot even when they are not used
        for qubit in qubits.values():
            if qubit.flux:
                self.config.register_flux_element(qubit)

        if not sequence:
            return {}
        # Current driver cannot play overlapping pulses on drive and flux channels
        # If we want to play overlapping pulses we need to define different elements on the same ports
        # like we do for readout multiplex
        qmsequence = QMSequence()
        for pulse in sequence:
            qmpulse = qmsequence.add(pulse)
            if pulse.duration % 4 or pulse.duration < 16:
                if pulse.type.name != "FLUX":
                    raise_error(NotImplementedError, "1ns resolution is available for flux pulses only.")
                # register flux element (if it does not already exist)
                self.config.register_flux_element(qubits[pulse.qubit], pulse.frequency)
                qmpulse.bake(self.config)
            else:
                self.config.register_pulse(qubits[pulse.qubit], pulse, self.time_of_flight, self.smearing)

        # play pulses using QUA
        with program() as experiment:
            n = declare(int)
            for qmpulse in qmsequence.ro_pulses:
                threshold = qubits[qmpulse.pulse.qubit].threshold
                iq_angle = qubits[qmpulse.pulse.qubit].iq_angle
                qmpulse.declare_output(threshold, iq_angle)

            with for_(n, 0, n < nshots, n + 1):
                self.play_pulses(qmsequence, relaxation_time)

            with stream_processing():
                # I_st.average().save("I")
                # Q_st.average().save("Q")
                # n_st.buffer().save_all("n")
                for qmpulse in qmsequence.ro_pulses:
                    serial = qmpulse.pulse.serial
                    qmpulse.I_st.buffer(nshots).save(f"{serial}_I")
                    qmpulse.Q_st.buffer(nshots).save(f"{serial}_Q")
                    if qmpulse.threshold is not None:
                        qmpulse.shots.buffer(nshots).save(f"{serial}_shots")

        result = self.execute_program(experiment)
        return self.fetch_results(result, sequence.ro_pulses)

    def sweep(self, qubits, sequence, *sweepers, nshots, relaxation_time, average=True):
        qmsequence = QMSequence()
        for pulse in sequence:
            qmpulse = qmsequence.add(pulse)
            if pulse.duration % 4 or pulse.duration < 16:
                if pulse.type.name != "FLUX":
                    raise_error(NotImplementedError, "1ns resolution is available for flux pulses only.")
                # register flux element (if it does not already exist)
                self.config.register_flux_element(qubits[pulse.qubit], pulse.frequency)
                qmpulse.bake(self.config)
            else:
                self.config.register_pulse(qubits[pulse.qubit], pulse, self.time_of_flight, self.smearing)

        # play pulses using QUA
        with program() as experiment:
            n = declare(int)
            for qmpulse in qmsequence.ro_pulses:
                if average:
                    # not calculating single shots when averaging
                    # during sweep so we do not pass ``threshold`` here
                    qmpulse.declare_output()
                else:
                    threshold = qubits[qmpulse.pulse.qubit].threshold
                    iq_angle = qubits[qmpulse.pulse.qubit].iq_angle
                    qmpulse.declare_output(threshold, iq_angle)

            with for_(n, 0, n < nshots, n + 1):
                self.sweep_recursion(list(sweepers), qubits, qmsequence, relaxation_time)

            with stream_processing():
                for qmpulse in qmsequence.ro_pulses:
                    Ist_temp = qmpulse.I_st
                    Qst_temp = qmpulse.Q_st
                    if not average and qmpulse.threshold is not None:
                        shots_temp = qmpulse.shots
                    for sweeper in reversed(sweepers):
                        Ist_temp = Ist_temp.buffer(len(sweeper.values))
                        Qst_temp = Qst_temp.buffer(len(sweeper.values))
                        if not average and qmpulse.threshold is not None:
                            shots_temp = shots_temp.buffer(len(sweeper.values))
                    serial = qmpulse.pulse.serial
                    if average:
                        Ist_temp.average().save(f"{serial}_I")
                        Qst_temp.average().save(f"{serial}_Q")
                    else:
                        Ist_temp.buffer(nshots).save(f"{serial}_I")
                        Qst_temp.buffer(nshots).save(f"{serial}_Q")
                        if qmpulse.threshold is not None:
                            shots_temp.buffer(nshots).save(f"{serial}_shots")

        result = self.execute_program(experiment)
        return self.fetch_results(result, sequence.ro_pulses)

    def sweep_frequency(self, sweepers, qubits, qmsequence, relaxation_time):
        from qm.qua import update_frequency

        sweeper = sweepers[0]
        freqs0 = []
        for pulse in sweeper.pulses:
            qubit = qubits[pulse.qubit]
            if pulse.type is PulseType.DRIVE:
                lo_frequency = int(qubit.drive.local_oscillator.frequency)
            elif pulse.type is PulseType.READOUT:
                lo_frequency = int(qubit.readout.local_oscillator.frequency)
            else:
                raise_error(NotImplementedError, f"Cannot sweep frequency of pulse of type {pulse.type}.")
            # convert to IF frequency for readout and drive pulses
            freqs0.append(declare(int, value=int(pulse.frequency - lo_frequency)))

        # is it fine to have this declaration inside the ``nshots`` QUA loop?
        f = declare(int)
        with for_(*from_array(f, sweeper.values.astype(int))):
            for pulse, f0 in zip(sweeper.pulses, freqs0):
                qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
                update_frequency(qmpulse.element, f + f0)

            self.sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)
            if relaxation_time > 0:
                wait(relaxation_time // 4)

    def sweep_amplitude(self, sweepers, qubits, qmsequence, relaxation_time):
        from qm.qua import amp

        # TODO: It should be -2 < amp(a) < 2 otherwise the we get weird results
        # without an error. Amplitude should be fixed to allow arbitrary values
        # in qibocal

        sweeper = sweepers[0]
        a = declare(fixed)
        with for_(*from_array(a, sweeper.values)):
            for pulse in sweeper.pulses:
                qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
                if qmpulse.baked is None:
                    qmpulse.operation = qmpulse.operation * amp(a)
                else:
                    qmpulse.baked_amplitude = a

            self.sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)
            if relaxation_time > 0:
                wait(relaxation_time // 4)

    def sweep_relative_phase(self, sweepers, qubits, qmsequence, relaxation_time):
        sweeper = sweepers[0]
        relphase = declare(fixed)
        with for_(*from_array(relphase, sweeper.values / (2 * np.pi))):
            for pulse in sweeper.pulses:
                qmpulse = qmsequence.pulse_to_qmpulse[pulse.serial]
                qmpulse.relative_phase = relphase

            self.sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)
            if relaxation_time > 0:
                wait(relaxation_time // 4)

    def sweep_bias(self, sweepers, qubits, qmsequence, relaxation_time):
        from qm.qua import set_dc_offset

        sweeper = sweepers[0]
        bias0 = [declare(fixed, value=qubits[q].flux.bias) for q in sweeper.qubits]
        b = declare(fixed)
        with for_(*from_array(b, sweeper.values)):
            for q, b0 in zip(sweeper.qubits, bias0):
                set_dc_offset(f"flux{q}", "single", b + b0)

            self.sweep_recursion(sweepers[1:], qubits, qmsequence, relaxation_time)
            if relaxation_time > 0:
                wait(relaxation_time // 4)

    SWEEPERS = {
        Parameter.frequency: sweep_frequency,
        Parameter.amplitude: sweep_amplitude,
        Parameter.relative_phase: sweep_relative_phase,
        Parameter.bias: sweep_bias,
    }

    def sweep_recursion(self, sweepers, qubits, qmsequence, relaxation_time):
        if len(sweepers) > 0:
            parameter = sweepers[0].parameter
            if parameter in self.SWEEPERS:
                self.SWEEPERS[parameter](self, sweepers, qubits, qmsequence, relaxation_time)
            else:
                raise_error(NotImplementedError, f"Sweeper for {parameter} is not implemented.")
        else:
            self.play_pulses(qmsequence)



from dataclasses import dataclass
from enum import Enum, auto
from typing import Optional

import numpy.typing as npt


class Parameter(Enum):
    """Sweeping parameters."""

    frequency = auto()
    amplitude = auto()
    relative_phase = auto()

    attenuation = auto()
    gain = auto()
    bias = auto()


QubitParameter = {Parameter.bias, Parameter.attenuation, Parameter.gain}


@dataclass
class Sweeper:
    """Data structure for Sweeper object.

    This object is passed as an argument to the method :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`
    which enables the user to sweep a specific parameter for one or more pulses. For information on how to
    perform sweeps see :func:`qibolab.platforms.abstract.AbstractPlatform.sweep`.

    Example:
        .. testcode::

            import numpy as np
            from qibolab.platform import Platform
            from qibolab.sweeper import Sweeper, Parameter
            from qibolab.pulses import PulseSequence


            platform = Platform("dummy")
            sequence = PulseSequence()
            parameter = Parameter.frequency
            pulse = platform.create_qubit_readout_pulse(qubit=0, start=0)
            sequence.add(pulse)
            parameter_range = np.random.randint(10, size=10)
            sweeper = Sweeper(parameter, parameter_range, [pulse])
            platform.sweep(sequence, sweeper)

    Args:
        parameter (`qibolab.sweeper.Parameter`): parameter to be swept, possible choices are frequency, attenuation, amplitude, current and gain.
        values (np.ndarray): sweep range. If the parameter is `frequency` the sweep will be a shift around the readout frequency
            in case of a `ReadoutPulse` or around the drive frequency for a generic `Pulse`. If the parameter is `amplitude` the range is
            normalized with the current amplitude of the pulse. For other parameters the sweep will be performed directly over the range specified.
        pulses (list) : list of `qibolab.pulses.Pulse` to be swept (optional).
        qubits (list): list of `qibolab.platforms.abstract.Qubit` to be swept (optional).
    """

    parameter: Parameter
    values: npt.NDArray
    pulses: Optional[list] = None
    qubits: Optional[list] = None

    def __post_init__(self):
        if self.pulses is not None and self.qubits is not None:
            raise ValueError("Cannot use a sweeper on both pulses and qubits.")
        elif self.pulses is not None and self.parameter in QubitParameter:
            raise ValueError(f"Cannot sweep {self.parameter} without specifying qubits.")
        elif self.qubits is not None and self.parameter not in QubitParameter:
            raise ValueError(f"Cannot sweep {self.parameter} without specifying pulses.")
        elif self.pulses is None and self.qubits is None:
            raise ValueError("Cannot use a sweeper without specifying pulses or qubits.")



from qibolab.instruments.abstract import LocalOscillator


class DummyLocalOscillator(LocalOscillator):
    """Dummy local oscillator driver.

    Useful for using with the Quantum Machines simulator.
    """

    def __init__(self, name, address):
        super().__init__(name, address)
        self.power: float
        self.frequency: float

    def connect(self):
        self.is_connected = True

    def setup(self, power=None, frequency=None, **kwargs):
        if power is not None:
            self.power = power
        if frequency is not None:
            self.frequency = frequency

    def start(self):
        pass

    def stop(self):
        pass

    def disconnect(self):
        self.is_connected = False



from dataclasses import dataclass, field
from typing import List

from qibo.config import log, raise_error

from qibolab.designs.channels import ChannelMap
from qibolab.instruments.abstract import AbstractInstrument


@dataclass
class InstrumentDesign:
    """Instrument design that uses a single controller.

    Attributes:
        instruments (list): List of :class:`qibolab.instruments.abstract.AbstractInstrument` objects.
        channels (:class:`qibolab.designs.ChannelMap`): Channel map describing the connections in the lab.
        _is_connected (bool): Boolean that shows whether instruments are connected.
    """

    instruments: List[AbstractInstrument]
    channels: ChannelMap = field(default_factory=ChannelMap)
    _is_connected: bool = field(default=False, init=False)

    def connect(self):
        """Connect to all instruments."""
        if not self._is_connected:
            for instrument in self.instruments:
                try:
                    log.info(f"Connecting to instrument {instrument}.")
                    instrument.connect()
                except Exception as exception:
                    raise_error(
                        RuntimeError,
                        f"Cannot establish connection to {instrument} instruments. Error captured: '{exception}'",
                    )
        self._is_connected = True

    def setup(self):
        """Load settings to instruments."""
        for instrument in self.instruments:
            instrument.setup()

    def start(self):
        """Start all instruments."""
        if self._is_connected:
            for instrument in self.instruments:
                instrument.start()

    def stop(self):
        """Stop all instruments."""
        if self._is_connected:
            for instrument in self.instruments:
                instrument.stop()

    def disconnect(self):
        """Disconnect all instruments."""
        if self._is_connected:
            for instrument in self.instruments:
                instrument.disconnect()
        self._is_connected = False

    def play(self, *args, **kwargs):
        """Play a pulse sequence and retrieve feedback."""
        result = {}
        for instrument in self.instruments:
            new_result = instrument.play(*args, **kwargs)
            if isinstance(new_result, dict):
                result.update(new_result)
            elif new_result is not None:
                # currently the result of QMSim is not a dict
                result = new_result
        return result

    def sweep(self, *args, **kwargs):
        """Play a pulse sequence while sweeping one or more parameters."""
        result = {}
        for instrument in self.instruments:
            new_result = instrument.sweep(*args, **kwargs)
            if isinstance(new_result, dict):
                result.update(new_result)
            elif new_result is not None:
                # currently the result of QMSim is not a dict
                result = new_result
        return result



import copy

from qibolab.platforms.multiqubit import MultiqubitPlatform


class Qubit:
    """Describes a single qubit in pulse control and readout extraction.

    Args:
        pi_pulse (dict): Qubit pi-pulse parameters.
            See qibolab.pulses.Pulse for more information.
        readout_pulse (dict): Qubit readout pulse parameters.
            See qibolab.pulses.ReadoutPulse for more information.
        resonator_spectroscopy_max_ro_voltage (float): Readout voltage corresponding to the ground state of the qubit.
        rabi_oscillations_pi_pulse_min_voltage (float): Readout voltage corresponding to the excited state of the qubit.
        playback (str): Instrument name for playing the qubit XY control pulses.
        playback_readout (str): Instrument name for playing the qubit readout pulse.
        readout_frequency (float): Readout frequency for IQ demodulation.
        readout (str): Instrument name for reading the qubit.
        readout_channels (int, int[]): Channels on the instrument associated to qubit readout.
    """

    def __init__(
        self,
        pi_pulse,
        readout_pulse,
        readout_frequency,
        resonator_spectroscopy_max_ro_voltage,
        rabi_oscillations_pi_pulse_min_voltage,
        playback,
        playback_readout,
        readout,
        readout_channels,
    ):
        self.pi_pulse = pi_pulse
        self.readout_pulse = readout_pulse
        self.readout_frequency = readout_frequency
        self.max_readout_voltage = resonator_spectroscopy_max_ro_voltage
        self.min_readout_voltage = rabi_oscillations_pi_pulse_min_voltage
        self.playback = playback
        self.playback_readout = playback_readout
        self.readout = readout
        self.readout_channels = readout_channels


class ICPlatform(MultiqubitPlatform):
    """Platform for controlling quantum devices with IC.

    Example:
        .. code-block:: python

            from qibolab import Platform

            platform = Platform("icarusq")

    """

    def __init__(self, name, runcard):
        self._instruments = []
        self._lo = []
        self._adc = []
        self._last_sequence = None
        super().__init__(name, runcard)
        self.qubits = []
        qubits = self.settings.get("qubits")
        for qubit_dict in qubits.values():
            self.qubits.append(Qubit(**qubit_dict))

    def run_calibration(self):  # pragma: no cover
        from qibo.config import raise_error

        raise_error(NotImplementedError)

    def execute_pulse_sequence(self, sequence, nshots=None):
        """Executes a pulse sequence. Pulses are being cached so that are not reuploaded
            if they are the same as the ones sent previously. This greatly accelerates
            some characterization routines that recurrently use the same set of pulses,
            i.e. qubit and resonator spectroscopy, spin echo, and future circuits based on
            fixed gates.

        Args:
            sequence (:class:`qibolab.pulses.PulseSequence`): Pulse sequence to execute.
            nshots (int): Number of shots to sample from the experiment.
                If ``None`` the default value provided as hardware_avg in the
                calibration json will be used.

        Returns:
            Readout results acquired by the assigned readout instrument
            after execution.
        """
        if not self.is_connected:
            from qibo.config import raise_error

            raise_error(RuntimeError, "Execution failed because instruments are not connected.")
        if nshots is None:
            nshots = self.hardware_avg

        from qibolab.pulses import ReadoutPulse

        qubits_to_measure = []
        measurement_results = []
        pulse_mapping = {}
        seq_serial = {}

        for pulse in sequence.pulses:
            # Assign pulses to each respective waveform generator
            qubit = self.fetch_qubit(pulse.qubit)
            playback_device = qubit.playback

            # Track each qubit to measure
            if isinstance(pulse, ReadoutPulse):
                qubits_to_measure.append(pulse.qubit)
                playback_device = qubit.playback_readout

            if playback_device not in pulse_mapping.keys():
                pulse_mapping[playback_device] = []
                seq_serial[playback_device] = []
            # Map the pulse to the associated playback instrument.
            pulse_mapping[playback_device].append(pulse)
            seq_serial[playback_device].append(pulse.serial)

        # Translate and upload the pulse subsequence for each device if needed
        for device, subsequence in pulse_mapping.items():
            inst = self.fetch_instrument(device)
            if self._last_sequence is None or seq_serial[device] != self._last_sequence[device]:
                inst.upload(inst.translate(subsequence, nshots))
            inst.play_sequence()
        self._last_sequence = seq_serial

        for adc in self._adc:
            adc.arm(nshots)

        # Start the experiment sequence
        self.start_experiment()

        # Fetch the experiment results
        for qubit_id in set(qubits_to_measure):
            qubit = self.fetch_qubit(qubit_id)
            inst = self.fetch_instrument(qubit.readout)
            measurement_results.append(inst.result(qubit.readout_frequency, qubit.readout_channels))

        if len(qubits_to_measure) == 1:
            return measurement_results[0]
        return measurement_results

    def fetch_instrument(self, name):
        """Returns a reference to an instrument."""
        try:
            res = next(inst for inst in self._instruments if inst.name == name)
            return res
        except StopIteration:
            from qibo.config import raise_error

            raise_error(Exception, "Instrument not found")

    def fetch_qubit(self, qubit_id=0) -> Qubit:
        """Fetches the qubit based on the id."""
        return self.qubits[qubit_id]

    def start_experiment(self):
        """Starts the instrument to start the experiment sequence."""
        inst = self.fetch_instrument(self.settings.get("settings").get("experiment_start_instrument"))
        inst.start_experiment()

    def fetch_qubit_pi_pulse(self, qubit_id=0) -> dict:
        """Fetches the qubit pi-pulse."""
        # Use copy to avoid mutability
        return copy.copy(self.fetch_qubit(qubit_id).pi_pulse)

    def fetch_qubit_readout_pulse(self, qubit_id=0) -> dict:
        """Fetches the qubit readout pulse."""
        # Use copy to avoid mutability
        return copy.copy(self.fetch_qubit(qubit_id).readout_pulse)



from qibo import gates
from qibo.config import log

from qibolab.transpilers.connectivity import fix_connectivity
from qibolab.transpilers.gate_decompositions import translate_gate


def transpile(circuit, two_qubit_natives, fuse_one_qubit=False, middle_qubit=2):
    """Implements full transpilation pipeline.

    Args:
        circuit (qibo.models.Circuit): Circuit model to transpile.
        two_qubit_natives (list): List of two qubit native gates
            supported by the quantum hardware ("CZ" and/or "iSWAP").
        fuse_one_qubit (bool): If ``True`` it fuses adjacent one-qubit gates to reduce
            circuit depth.
        middle_qubit (int): Hardware middle qubit.

    Returns:
        new (qibo.models.Circuit): New circuit that can be executed on tii5q platform.
        hardware_qubits (dict): Mapping between logical and physical qubits.
    """
    # Re-arrange gates using qibo's fusion algorithm
    # this may reduce number of SWAPs when fixing for connectivity
    fcircuit = circuit.fuse(max_qubits=2)
    new = circuit.__class__(circuit.nqubits)
    for fgate in fcircuit.queue:
        if isinstance(fgate, gates.FusedGate):
            new.add(fgate.gates)
        else:
            new.add(fgate)

    # Add SWAPs to satisfy connectivity constraints
    new, hardware_qubits = fix_connectivity(circuit, middle_qubit=middle_qubit)

    # two-qubit gates to native
    new = translate_circuit(new, two_qubit_natives, translate_single_qubit=False)

    # Optional: fuse one-qubit gates to reduce circuit depth
    if fuse_one_qubit:
        new = new.fuse(max_qubits=1)

    # one-qubit gates to native
    new = translate_circuit(new, two_qubit_natives, translate_single_qubit=True)

    return new, hardware_qubits


def translate_circuit(circuit, two_qubit_natives, translate_single_qubit=False):
    """Translates a circuit to native gates.

    Args:
        circuit (qibo.models.Circuit): Circuit model to translate into native gates.
        two_qubit_natives (list): List of two qubit native gates
            supported by the quantum hardware ("CZ" and/or "iSWAP").

    Returns:
        new (qibo.models.Circuit): Equivalent circuit with native gates.
    """
    new = circuit.__class__(circuit.nqubits)
    for gate in circuit.queue:
        if len(gate.qubits) > 1 or translate_single_qubit:
            new.add(translate_gate(gate, two_qubit_natives))
        else:
            new.add(gate)
    return new


def can_execute(circuit, two_qubit_natives, middle_qubit=2, verbose=True):
    """Checks if a circuit can be executed on tii5q.

    Args:
        circuit (qibo.models.Circuit): Circuit model to check.
        two_qubit_natives (list): List of two qubit native gates
            supported by the quantum hardware ("CZ" and/or "iSWAP").
        middle_qubit (int): Hardware middle qubit.
        verbose (bool): If ``True`` it prints debugging log messages.

    Returns ``True`` if the following conditions are satisfied:
        - Circuit does not contain more than two-qubit gates.
        - All one-qubit gates are I, Z, RZ or U3.
        - All two-qubit gates are CZ or iSWAP based on two_qubit_natives.
        - All two-qubit gates have qubit 0 as target or control.
    otherwise returns ``False``.
    """

    # pring messages only if ``verbose == True``
    vlog = lambda msg: log.info(msg) if verbose else lambda msg: None
    for gate in circuit.queue:
        if isinstance(gate, gates.M):
            continue

        if len(gate.qubits) == 1:
            if not isinstance(gate, (gates.I, gates.Z, gates.RZ, gates.U3)):
                vlog(f"{gate.name} is not a single qubit native gate.")
                return False

        elif len(gate.qubits) == 2:
            if gate.__class__.__name__ not in two_qubit_natives:
                vlog(f"{gate.name} is not a two qubit native gate.")
                return False
            if middle_qubit not in gate.qubits:
                vlog("Circuit does not respect connectivity. " f"{gate.name} acts on {gate.qubits}.")
                return False

        else:
            vlog(f"{gate.name} acts on more than two qubits.")
            return False

    vlog("Circuit can be executed.")
    return True



import random
from enum import Enum, auto

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from more_itertools import pairwise
from qibo import gates
from qibo.config import log, raise_error
from qibo.models import Circuit

DEFAULT_INIT_SAMPLES = 100


class QubitInitMethod(Enum):
    """A class to define the initial qubit mapping methods."""

    greedy = auto()
    subgraph = auto()
    custom = auto()


class Transpiler:
    """A class to perform initial qubit mapping and connectivity matching.

    Properties:
        connectivity (networkx.graph): chip connectivity.
        init_method (string or QubitInitMethod): initial qubit mapping method.
        init_samples (int): number of random qubit initializations for greedy initial qubit mapping.

    Attributes:
        _circuit_repr (list): quantum circuit represented as a list (only 2 qubit gates).
        _mapping (dict): circuit to physical qubit mapping during transpiling.
        _graph (networkx.graph): qubit mapped as nodes of the connectivity graph.
        _qubit_map (np.array): circuit to physical qubit mapping during transpiling as vector.
        _circuit_position (int): position in the circuit.
        _added_swaps (int): number of swaps added to the circuit to match connectivity.
    """

    def __init__(self, connectivity, init_method="greedy", init_samples=None):
        self.connectivity = connectivity
        self.init_method = init_method
        if self.init_method is QubitInitMethod.greedy and init_samples is None:
            init_samples = DEFAULT_INIT_SAMPLES
        self.init_samples = init_samples

        self._circuit_repr = None
        self._mapping = None
        self._graph = None
        self._qubit_map = None
        self._circuit_position = 0
        self._added_swaps = 0

    def transpile(self, qibo_circuit):
        """Qubit mapping initialization and circuit transpiling.

        Args:
            qibo_circuit (qibo.Circuit): circuit to be transpiled.

        Returns:
            hardware_mapped_circuit (qibo.Circuit): circut mapped to hardware topology.
            final_mapping (dict): final qubit mapping.
            init_mapping (dict): initial qubit mapping.
            added_swaps (int): number of swap gates added.
        """
        self._circuit_position = 0
        self._added_swaps = 0
        self.translate_circuit(qibo_circuit)
        keys = list(self._connectivity.nodes())
        if self._init_method is QubitInitMethod.greedy:
            self.greedy_init()
        elif self._init_method is QubitInitMethod.subgraph:
            self.subgraph_init()
        elif self._init_method is QubitInitMethod.custom:
            self._mapping = dict(zip(keys, self._mapping.values()))
            self._graph = nx.relabel_nodes(self._connectivity, self._mapping)
        # Inverse permutation
        init_qubit_map = np.argsort(list(self._mapping.values()))
        init_mapping = dict(zip(keys, init_qubit_map))
        self._qubit_map = np.sort(init_qubit_map)
        self.init_circuit(qibo_circuit)
        self.first_transpiler_step(qibo_circuit)
        while len(self._circuit_repr) != 0:
            self.transpiler_step(qibo_circuit)
        final_mapping = {key: init_qubit_map[self._qubit_map[i]] for i, key in enumerate(keys)}
        return (
            self.init_mapping_circuit(self.transpiled_circuit, init_qubit_map),
            final_mapping,
            init_mapping,
            self._added_swaps,
        )

    def transpiler_step(self, qibo_circuit):
        """Transpilation step. Find new mapping, add swap gates and apply gates that can be run with this configuration.

        Args:
            qibo_circuit (qibo.Circuit): circuit to be transpiled.
        """
        len_2q_circuit = len(self._circuit_repr)
        path, meeting_point = self.relocate()
        self.add_swaps(path, meeting_point)
        self.update_qubit_map()
        self.add_gates(qibo_circuit, len_2q_circuit - len(self._circuit_repr))

    def first_transpiler_step(self, qibo_circuit):
        """First transpilation step. Apply gates that can be run with the initial qubit mapping.

        Args:
            qibo_circuit (qibo.Circuit): circuit to be transpiled.
        """
        len_2q_circuit = len(self._circuit_repr)
        self._circuit_repr = self.reduce(self._graph)
        self.add_gates(qibo_circuit, len_2q_circuit - len(self._circuit_repr))

    def custom_qubit_mapping(self, map):
        """Define a custom initial qubit mapping.

        Args:
            map (list): List reporting the circuit to chip qubit mapping,
            example [1,2,0] to assign the logical to physical qubit mapping.
        """
        self.init_method = QubitInitMethod.custom
        self._mapping = dict(zip(range(len(map)), map))

    @property
    def connectivity(self):
        return self._connectivity

    @connectivity.setter
    def connectivity(self, connectivity):
        """Set the hardware chip connectivity.

        Args:
            connectivity (networkx graph): define connectivity.
        """

        if isinstance(connectivity, nx.Graph):
            self._connectivity = connectivity
        else:
            raise_error(TypeError, "Use networkx graph for custom connectivity")

    def draw_connectivity(self):  # pragma: no cover
        """Show connectivity graph."""
        pos = nx.spectral_layout(self._connectivity)
        nx.draw(self._connectivity, pos=pos, with_labels=True)
        plt.show()

    @property
    def init_method(self):
        return self._init_method

    @init_method.setter
    def init_method(self, init_method, init_samples=100):
        """Set the initial mapping method for the transpiler.

        Args:
            init_method (string): Initial mapping method ("greedy" or "subgraph").
        """
        if isinstance(init_method, str):
            init_method = QubitInitMethod[init_method]
        self._init_method = init_method

    @property
    def init_samples(self):
        return self._init_samples

    @init_samples.setter
    def init_samples(self, init_samples):
        """Set the initial mapping method for the transpiler.

        Args:
            init_samples (int): Number of random qubit mapping samples to be tested
            (required only for "greedy" initialization).
        """
        if init_samples is not None:
            self.init_method = QubitInitMethod.greedy
        self._init_samples = init_samples

    def translate_circuit(self, qibo_circuit):
        """Translate qibo circuit into a list of two qubit gates to be used by the transpiler.

        Args:
            qibo_circuit (qibo.Circuit): circuit to be transpiled.
        """
        translated_circuit = []
        for index, gate in enumerate(qibo_circuit.queue):
            if len(gate.qubits) == 2:
                gate_qubits = list(gate.qubits)
                gate_qubits.sort()
                gate_qubits.append(index)
                translated_circuit.append(gate_qubits)
            if len(gate.qubits) >= 3:
                raise_error(ValueError, "ERROR do not use gates acting on more than 2 qubits")
        self._circuit_repr = translated_circuit

    def reduce(self, graph):
        """Reduce the circuit, delete a 2-qubit gate if it can be applied on the current configuration.

        Args:
            graph (networkx.Graph): current hardware qubit mapping.

        Returns:
            new_circuit (list): reduced circuit.
        """
        new_circuit = self._circuit_repr.copy()
        while new_circuit != [] and (new_circuit[0][0], new_circuit[0][1]) in graph.edges():
            del new_circuit[0]
        return new_circuit

    def subgraph_init(self):
        """Subgraph isomorphism initialization, NP-complete it can take a long time for large circuits."""
        H = nx.Graph()
        H.add_nodes_from([i for i in range(self._connectivity.number_of_nodes())])
        GM = nx.algorithms.isomorphism.GraphMatcher(self._connectivity, H)
        i = 0
        H.add_edge(self._circuit_repr[i][0], self._circuit_repr[i][1])
        while GM.subgraph_is_monomorphic() == True:
            result = GM
            i = i + 1
            H.add_edge(self._circuit_repr[i][0], self._circuit_repr[i][1])
            GM = nx.algorithms.isomorphism.GraphMatcher(self._connectivity, H)
            if self._connectivity.number_of_edges() == H.number_of_edges() or i == len(self._circuit_repr) - 1:
                G = nx.relabel_nodes(self._connectivity, result.mapping)
                self._graph = G
                self._mapping = result.mapping
                return
        G = nx.relabel_nodes(self._connectivity, result.mapping)
        self._graph = G
        self._mapping = result.mapping

    def greedy_init(self):
        """Initialize the circuit with greedy algorithm let a maximum number of 2-qubit
        gates can be applied without introducing any SWAP gate"""
        nodes = self._connectivity.number_of_nodes()
        keys = list(self._connectivity.nodes())
        final_mapping = dict(zip(keys, list(range(nodes))))
        final_graph = nx.relabel_nodes(self._connectivity, final_mapping)
        final_cost = len(self.reduce(final_graph))
        for _ in range(self._init_samples):
            mapping = dict(zip(keys, random.sample(range(nodes), nodes)))
            graph = nx.relabel_nodes(self._connectivity, mapping)
            cost = len(self.reduce(graph))
            if cost == 0:
                self._graph = graph
                self._mapping = mapping
                return
            if cost < final_cost:
                final_graph = graph
                final_mapping = mapping
                final_cost = cost
        self._graph = final_graph
        self._mapping = final_mapping

    def map_list(self, path):
        """Return all possible walks of qubits for a given path.

        Args:
            path (list): path to move qubits.

        Returns:
            mapping_list (list): all possible walks of qubits for a given path.
            meeting_point_list (list): all possible qubit meeting point in the path.
        """
        path_ends = [path[0], path[-1]]
        path_middle = path[1:-1]
        mapping_list = []
        meeting_point_list = []
        for i in range(len(path) - 1):
            values = path_middle[:i] + path_ends + path_middle[i:]
            mapping = dict(zip(path, values))
            mapping_list.append(mapping)
            meeting_point_list.append(i)
        return mapping_list, meeting_point_list

    def relocate(self):
        """A small greedy algorithm to decide which path to take, and how qubits should walk.

        Returns:
            final_path (list): best path to move qubits.
            meeting_point (int): qubit meeting point in the path.
        """
        nodes = self._graph.number_of_nodes()
        circuit = self.reduce(self._graph)
        final_circuit = circuit
        keys = list(range(nodes))
        final_graph = self._graph
        final_mapping = dict(zip(keys, keys))
        # Consider all shortest paths
        path_list = [p for p in nx.all_shortest_paths(self._graph, source=circuit[0][0], target=circuit[0][1])]
        self._added_swaps += len(path_list[0]) - 2
        final_path = path_list[0]
        # Reduce the number of paths to be faster
        for path in path_list:
            list_, meeting_point_list = self.map_list(path)
            for j, mapping in enumerate(list_):
                new_graph = nx.relabel_nodes(self._graph, mapping)
                new_circuit = self.reduce(new_graph)
                # Greedy looking for the optimal path and the optimal walk on this path
                if len(new_circuit) < len(final_circuit):
                    final_graph = new_graph
                    final_circuit = new_circuit
                    final_mapping = mapping
                    final_path = path
                    meeting_point = meeting_point_list[j]
        self._graph = final_graph
        self._mapping = final_mapping
        self._circuit_repr = final_circuit
        return final_path, meeting_point

    def init_circuit(self, qibo_circuit):
        """Initialize the transpiled circuit

        Args:
            Args: qibo_circuit (qibo.Circuit): circuit to be transpiled.
        """
        nodes = self._connectivity.number_of_nodes()
        qubits = qibo_circuit.nqubits
        if qubits > nodes:
            raise_error(ValueError, "There are not enough physical qubits in the hardware to map the circuit")
        elif qubits == nodes:
            new_circuit = Circuit(nodes)
        else:
            log.warning(
                "You are using more physical qubits than required by the circuit, some qubits will be added to the circuit"
            )
            new_circuit = Circuit(nodes)
        self.transpiled_circuit = new_circuit

    def init_mapping_circuit(self, circuit, qubit_map):
        """Initial qubit mapping of the transpiled qibo circuit

        Args:
            circuit (qibo.Circuit): transpiled qibo circuit.
            qubit_map (np.array): initial qubit mapping.

        Returns:
            new_circuit (qibo.Circuit): transpiled circuit mapped with initial qubit mapping.
        """
        new_circuit = Circuit(self._connectivity.number_of_nodes())
        for gate in circuit.queue:
            new_circuit.add(gate.on_qubits({q: qubit_map[q] for q in gate.qubits}))
        return new_circuit

    def add_gates(self, qibo_circuit, matched_gates):
        """Add one and two qubit gates to transpiled circuit until connectivity is matched

        Args:
            qibo_circuit (qibo.Circuit): circuit to be transpiled.
            matched_gates (int): number of two qubit gates that can be applied with the current qubit mapping
        """
        index = 0
        while self._circuit_position < len(qibo_circuit.queue):
            gate = qibo_circuit.queue[self._circuit_position]
            if len(gate.qubits) == 1:
                self.transpiled_circuit.add(gate.on_qubits({gate.qubits[0]: self._qubit_map[gate.qubits[0]]}))
                self._circuit_position += 1
            else:
                index += 1
                if index == matched_gates + 1:
                    break
                else:
                    self.transpiled_circuit.add(
                        gate.on_qubits(
                            {
                                gate.qubits[0]: self._qubit_map[gate.qubits[0]],
                                gate.qubits[1]: self._qubit_map[gate.qubits[1]],
                            }
                        )
                    )
                    self._circuit_position += 1

    def add_swaps(self, path, meeting_point):
        """Add swaps to the transpiled circuit to move qubits

        Args:
            path (list): path to move qubits.
            meeting_point (int): qubit meeting point in the path.
        """
        forward = path[0 : meeting_point + 1]
        backward = path[meeting_point + 1 :: -1]
        if len(forward) > 1:
            for f1, f2 in pairwise(forward):
                self.transpiled_circuit.add(gates.SWAP(self._qubit_map[f1], self._qubit_map[f2]))
        if len(backward) > 1:
            for b1, b2 in pairwise(backward):
                self.transpiled_circuit.add(gates.SWAP(self._qubit_map[b1], self._qubit_map[b2]))

    def update_qubit_map(self):
        """Update the qubit mapping after adding swaps"""
        old_mapping = self._qubit_map.copy()
        for key, value in self._mapping.items():
            self._qubit_map[value] = old_mapping[key]



import os
from pathlib import Path

qibolab_folder = Path(__file__).parent
src_folder = qibolab_folder.parent
project_folder = src_folder.parent
examples_folder = project_folder / "examples"
script_folder = Path(os.path.abspath(""))
user_folder = Path.home() / ".qibolab"



import networkx as nx
from qibo.config import raise_error

from qibolab.designs import Channel, ChannelMap, InstrumentDesign
from qibolab.platforms.platform import DesignPlatform


def create_dummy(runcard):
    """Create a dummy platform using the dummy instrument.

    Useful for testing.
    """
    from qibolab.instruments.dummy import DummyInstrument

    # Create channel objects
    channels = ChannelMap()
    channels |= ChannelMap.from_names("readout", "drive", "flux")

    # Create dummy controller
    instrument = DummyInstrument("dummy", 0)
    # Create design
    design = InstrumentDesign([instrument], channels)
    # Create platform
    platform = DesignPlatform("dummy", design, runcard)

    # map channels to qubits
    for qubit in platform.qubits:
        platform.qubits[qubit].readout = channels["readout"]
        platform.qubits[qubit].drive = channels["drive"]
        platform.qubits[qubit].flux = channels["flux"]

    return platform


def create_tii_qw5q_gold(runcard, simulation_duration=None, address=None, cloud=False):
    """Create platform using Quantum Machines (QM) OPXs and Rohde Schwarz local oscillators.

    IPs and other instrument related parameters are hardcoded in ``__init__`` and ``setup``.

    Args:
        runcard (str): Path to the runcard file.
        simulation_duration (int): Duration for the simulation in ns.
            If given the compiler simulator will be used instead of the actual hardware.
            Default is ``None`` which falls back to the hardware.
        address (str): Address and port for the QM OPX cluster.
            If ``None`` it will attempt to connect to TII instruments.
        cloud (bool): See :class:`qibolab.instruments.qmsim.QMSim` for details.
            Relevant only when ``simulation_duration`` is given.
    """
    # Create channel objects
    channels = ChannelMap()
    # readout
    channels |= ChannelMap.from_names("L3-25_a", "L3-25_b")
    # feedback
    channels |= ChannelMap.from_names("L2-5_a", "L2-5_b")
    # drive
    channels |= ChannelMap.from_names(*(f"L3-{i}" for i in range(11, 16)))
    # flux
    channels |= ChannelMap.from_names(*(f"L4-{i}" for i in range(1, 6)))
    # TWPA
    channels |= ChannelMap.from_names("L4-26")

    # Map controllers to qubit channels (HARDCODED)
    # readout
    channels["L3-25_a"].ports = [("con1", 10), ("con1", 9)]
    channels["L3-25_b"].ports = [("con2", 10), ("con2", 9)]
    # feedback
    channels["L2-5_a"].ports = [("con1", 2), ("con1", 1)]
    channels["L2-5_b"].ports = [("con2", 2), ("con2", 1)]
    # drive
    for i in range(1, 5):
        channels[f"L3-1{i}"].ports = [("con1", 2 * i), ("con1", 2 * i - 1)]
    channels["L3-15"].ports = [("con3", 2), ("con3", 1)]
    # flux
    for i in range(1, 6):
        channels[f"L4-{i}"].ports = [("con2", i)]

    # Instantiate QM OPX instruments
    if simulation_duration is None:
        from qibolab.instruments.qm import QMOPX
        from qibolab.instruments.rohde_schwarz import SGS100A as LocalOscillator

        controller = QMOPX("qmopx", "192.168.0.1:80")

    else:
        from qibolab.instruments.dummy_oscillator import (
            DummyLocalOscillator as LocalOscillator,
        )
        from qibolab.instruments.qmsim import QMSim

        if address is None:
            # connect to TII instruments for simulation
            address = "192.168.0.1:80"

        controller = QMSim("qmopx", address, simulation_duration, cloud)

    # set time of flight for readout integration (HARDCODED)
    controller.time_of_flight = 280

    # Instantiate local oscillators (HARDCODED)
    local_oscillators = [
        LocalOscillator("lo_readout_a", "192.168.0.39"),
        LocalOscillator("lo_readout_b", "192.168.0.31"),
        LocalOscillator("lo_drive_low", "192.168.0.32"),
        LocalOscillator("lo_drive_mid", "192.168.0.33"),
        LocalOscillator("lo_drive_high", "192.168.0.34"),
        LocalOscillator("twpa_a", "192.168.0.35"),
    ]
    # Set LO parameters
    local_oscillators[0].frequency = 7_300_000_000
    local_oscillators[1].frequency = 7_900_000_000
    local_oscillators[2].frequency = 4_700_000_000
    local_oscillators[3].frequency = 5_600_000_000
    local_oscillators[4].frequency = 6_500_000_000
    local_oscillators[0].power = 18.0
    local_oscillators[1].power = 15.0
    for i in range(2, 5):
        local_oscillators[i].power = 16.0
    # Set TWPA parameters
    local_oscillators[5].frequency = 6_511_000_000
    local_oscillators[5].power = 4.5
    # Map LOs to channels
    channels["L3-25_a"].local_oscillator = local_oscillators[0]
    channels["L3-25_b"].local_oscillator = local_oscillators[1]
    channels["L3-15"].local_oscillator = local_oscillators[2]
    channels["L3-11"].local_oscillator = local_oscillators[2]
    channels["L3-12"].local_oscillator = local_oscillators[3]
    channels["L3-13"].local_oscillator = local_oscillators[4]
    channels["L3-14"].local_oscillator = local_oscillators[4]
    channels["L4-26"].local_oscillator = local_oscillators[5]

    instruments = [controller] + local_oscillators
    design = InstrumentDesign(instruments, channels)
    platform = DesignPlatform("qw5q_gold", design, runcard)

    # assign channels to qubits
    qubits = platform.qubits
    for q in [0, 1, 5]:
        qubits[q].readout = channels["L3-25_a"]
        qubits[q].feedback = channels["L2-5_a"]
    for q in [2, 3, 4]:
        qubits[q].readout = channels["L3-25_b"]
        qubits[q].feedback = channels["L2-5_b"]

    qubits[0].drive = channels["L3-15"]
    qubits[0].flux = channels["L4-5"]
    channels["L4-5"].qubit = qubits[0]
    for q in range(1, 5):
        qubits[q].drive = channels[f"L3-{10 + q}"]
        qubits[q].flux = channels[f"L4-{q}"]
        channels[f"L4-{q}"].qubit = qubits[q]

    # Platfom topology
    Q = [f"q{i}" for i in range(5)]
    chip = nx.Graph()
    chip.add_nodes_from(Q)
    graph_list = [
        (Q[0], Q[2]),
        (Q[1], Q[2]),
        (Q[3], Q[2]),
        (Q[4], Q[2]),
    ]
    chip.add_edges_from(graph_list)
    platform.topology = chip

    return platform


def Platform(name, runcard=None, design=None):
    """Platform for controlling quantum devices.

    Args:
        name (str): name of the platform. Options are 'tiiq', 'qili' and 'icarusq'.
        runcard (str): path to the yaml file containing the platform setup.
        design (:class:`qibolab.designs.abstract.AbstractInstrumentDesign`): Instrument
            design to use for the platform.

    Returns:
        The plaform class.
    """
    if not runcard:
        from os.path import exists

        from qibolab.paths import qibolab_folder

        runcard = qibolab_folder / "runcards" / f"{name}.yml"
        if not exists(runcard):
            raise_error(RuntimeError, f"Runcard {name} does not exist.")

    if name == "dummy":
        return create_dummy(runcard)
    elif name == "icarusq":
        from qibolab.platforms.icplatform import ICPlatform as Device
    elif name == "qw5q_gold":
        return create_tii_qw5q_gold(runcard)
    else:
        from qibolab.platforms.multiqubit import MultiqubitPlatform as Device

    return Device(name, runcard)



from dataclasses import dataclass, field
from typing import List, Optional

from qibo.config import raise_error

from qibolab.instruments.abstract import LocalOscillator


@dataclass
class Channel:
    """Representation of physical wire connection (channel)."""

    name: str
    """Name of the channel from the lab schematics."""

    qubit: Optional["Qubit"] = field(default=None, repr=False)
    """Qubit connected to this channel.
    Used to read the sweetspot and filters for flux channels only. ``None`` for non-flux channels.
    """
    ports: List[tuple] = field(default_factory=list)
    """List of tuples (controller, port) connected to this channel."""
    local_oscillator: Optional[LocalOscillator] = None
    """Instrument object controlling the local oscillator connected to this channel.
    Not applicable for setups that do not use local oscillators because the controller
    can send sufficiently high frequencies
    """
    _bias: Optional[float] = None
    """DC offset that should be applied in the channel in order to shift the
    frequency of the qubit, usually to put it in its sweetspot.
    Relevant only for flux channels and flux-tunable transmon qubits.
    """
    _filter: Optional[dict] = None
    """Filter to be applied to the channel to reduce the distortions when sending
    flux pulses. Useful for two-qubit gates.
    Quantum Machines associate filters to channels but this may not be the case
    in other instruments.
    """

    @property
    def bias(self):
        """Bias offset for flux channels."""
        if self._bias is None:
            # operate qubits at their sweetspot unless otherwise stated
            return self.qubit.sweetspot
        return self._bias

    @bias.setter
    def bias(self, bias):
        if not isinstance(bias, (int, float)):
            raise_error(TypeError, f"Attempting to set non-float bias {bias}.")
        self._bias = bias

    @property
    def filter(self):
        """Filters for sending flux pulses through a flux channel."""
        if self._filter is None:
            return self.qubit.filter
        return self._filter

    @filter.setter
    def filter(self, filter):
        if not isinstance(filter, dict):
            raise_error(TypeError, f"Channel filter must be dict but is {type(filter)}.")
        self._filter = filter


@dataclass
class ChannelMap:
    """Collection of :class:`qibolab.designs.channel.Channel` objects identified by name."""

    _channels: dict = field(default_factory=dict)

    @classmethod
    def from_names(cls, *names):
        """Construct multiple :class:`qibolab.designs.channel.Channel` objects from names.

        Args:
            names (str): List of channel names.
        """
        return cls({name: Channel(name) for name in names})

    def __getitem__(self, name):
        return self._channels[name]

    def __setitem__(self, name, channel):
        if not isinstance(channel, Channel):
            raise_error(TypeError, f"Cannot add channel of type {type(channel)} to ChannelMap.")
        self._channels[name] = channel

    def __contains__(self, name):
        return name in self._channels

    def __or__(self, channel_map):
        channels = self._channels.copy()
        channels.update(channel_map._channels)
        return self.__class__(channels)

    def __ior__(self, channel_map):
        self._channels.update(channel_map._channels)
        return self
